/* ==========================================================================
 * delegate.h - Not quite closures.
 * --------------------------------------------------------------------------
 * Copyright (c) 2009  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#ifndef DELEGATE_H
#define DELEGATE_H

#include <stdint.h>

#include <limits.h>

#include <string.h>

#if __linux
#include <ffi.h>
#else
#include <ffi/ffi.h>
#endif


#define ffi_bits(T)	(sizeof (T) * CHAR_BIT)

#define ffi_int8_compatible(T) \
	((__builtin_types_compatible_p(T, int8_t)) || \
	 (__builtin_types_compatible_p(T, signed char) && ffi_bits(T) == 8))

#define ffi_uint8_compatible(T) \
	((__builtin_types_compatible_p(T, uint8_t)) || \
	 (__builtin_types_compatible_p(T, unsigned char) && ffi_bits(T) == 8))

#define ffi_int16_compatible(T) \
	((__builtin_types_compatible_p(T, int16_t)) || \
	 (__builtin_types_compatible_p(T, signed char) && ffi_bits(T) == 16) || \
	 (__builtin_types_compatible_p(T, short) && ffi_bits(T) == 16))

#define ffi_uint16_compatible(T) \
	((__builtin_types_compatible_p(T, uint16_t)) || \
	 (__builtin_types_compatible_p(T, unsigned char) && ffi_bits(T) == 16) || \
	 (__builtin_types_compatible_p(T, unsigned short) && ffi_bits(T) == 16))

#define ffi_int32_compatible(T) \
	((__builtin_types_compatible_p(T, int32_t)) || \
	 (__builtin_types_compatible_p(T, int) && ffi_bits(T) == 32) || \
	 (__builtin_types_compatible_p(T, long) && ffi_bits(T) == 32))

#define ffi_uint32_compatible(T) \
	((__builtin_types_compatible_p(T, uint32_t)) || \
	 (__builtin_types_compatible_p(T, unsigned int) && ffi_bits(T) == 32) || \
	 (__builtin_types_compatible_p(T, unsigned long) && ffi_bits(T) == 32))

#define ffi_int64_compatible(T) \
	((__builtin_types_compatible_p(T, int64_t)) || \
	 (__builtin_types_compatible_p(T, long) && ffi_bits(T) == 64) || \
	 (__builtin_types_compatible_p(T, long long) && ffi_bits(T) == 64))

#define ffi_uint64_compatible(T) \
	((__builtin_types_compatible_p(T, uint64_t)) || \
	 (__builtin_types_compatible_p(T, unsigned long) && ffi_bits(T) == 64) || \
	 (__builtin_types_compatible_p(T, unsigned long long) && ffi_bits(T) == 64))


#define ffi_typeof(E) __typeof__(					\
	__builtin_choose_expr(						\
	ffi_uint8_compatible(__typeof__(E)),				\
	(uint8_t){ 0 },							\
	__builtin_choose_expr(						\
	ffi_int8_compatible(__typeof__(E)),				\
	(int8_t){ 0 },							\
	__builtin_choose_expr(						\
	ffi_uint16_compatible(__typeof__(E)),				\
	(uint16_t){ 0 },						\
	__builtin_choose_expr(						\
	ffi_int16_compatible(__typeof__(E)),				\
	(int16_t){ 0 },							\
	__builtin_choose_expr(						\
	ffi_uint32_compatible(__typeof__(E)),				\
	(uint32_t){ 0 },						\
	__builtin_choose_expr(						\
	ffi_int32_compatible(__typeof__(E)),				\
	(int32_t){ 0 },							\
	__builtin_choose_expr(						\
	ffi_uint64_compatible(__typeof__(E)),				\
	(uint64_t){ 0 },						\
	__builtin_choose_expr(						\
	ffi_int64_compatible(__typeof__(E)),				\
	(int64_t){ 0 },							\
	__builtin_choose_expr(						\
	__builtin_types_compatible_p(__typeof__(E), float),		\
	(float){ 0 },							\
	__builtin_choose_expr(						\
	__builtin_types_compatible_p(__typeof__(E), double),		\
	(double){ 0 },							\
	__builtin_choose_expr(						\
	__builtin_types_compatible_p(__typeof__(E), long double),	\
	(long double){ 0 },						\
	__builtin_choose_expr(						\
	__builtin_types_compatible_p(__typeof__(E), void *),		\
	(void *){ 0 },							\
	__builtin_choose_expr(						\
	__builtin_types_compatible_p(__typeof__(E), void),		\
	(void)0, (void *){ 0 }						\
	))))))))))))))


#define ffi_type(T)					\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, uint8_t),	\
	&ffi_type_uint8,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, int8_t),	\
	&ffi_type_sint8,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, uint16_t),	\
	&ffi_type_uint16,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, int16_t),	\
	&ffi_type_sint16,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, uint32_t),	\
	&ffi_type_uint32,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, int32_t),	\
	&ffi_type_sint32,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, uint64_t),	\
	&ffi_type_uint64,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, int64_t),	\
	&ffi_type_sint64,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, float),		\
	&ffi_type_float,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, double),	\
	&ffi_type_double,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, void *),	\
	&ffi_type_pointer,				\
	__builtin_choose_expr(				\
	__builtin_types_compatible_p(T, void),		\
	&ffi_type_void, &ffi_type_void			\
	))))))))))))


#define D_D1 0
#define D_D2 1
#define D_D3 2
#define D_D4 3
#define D_D5 4
#define D_D6 5
#define D_D7 6
#define D_D8 7
#define D_D9 8
#define D_DEC(N) D_XPASTE(D_D, N)

#define D_ARG0(a, ...) a
#define D_ARG1(a, ...) D_ARG0(__VA_ARGS__)
#define D_ARG2(a, ...) D_ARG1(__VA_ARGS__)
#define D_ARG3(a, ...) D_ARG2(__VA_ARGS__)
#define D_ARG4(a, ...) D_ARG3(__VA_ARGS__)
#define D_ARG5(a, ...) D_ARG4(__VA_ARGS__)
#define D_ARG6(a, ...) D_ARG5(__VA_ARGS__)
#define D_ARG7(a, ...) D_ARG6(__VA_ARGS__)
#define D_ARG8(a, ...) D_ARG7(__VA_ARGS__)
#define D_ARG9(a, ...) D_ARG8(__VA_ARGS__)
#define D_ARG_(M, ...) M(__VA_ARGS__)
#define D_ARG(N, ...) D_ARG_(D_XPASTE(D_ARG, N), __VA_ARGS__)

#define D_NARG_(_15, _14, _13, _12, _11, _10, _9, _8, _7, _6, _5, _4, _3, _2, _1, N, ...) N
#define D_NARG(...) D_NARG_(__VA_ARGS__, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1)

#define D_CALL(M, ...) M(__VA_ARGS__)

#define D_PASTE(a, b) a ## b
#define D_XPASTE(a, b) D_PASTE(a, b)


#define DELEGATE_INITIALIZER { 0 }

struct delegate {
	void (*fn)();

	ffi_type *rtype;

	int argc;

	union {
		intmax_t i;
		void *p;
		long double g;
	} args[7];

	ffi_type *types[7];

	ffi_cif cif;
}; /* struct delegate */


#define d_copyarg(del, argn, T, arg) \
	({ __typeof__(T) argT; \
	   memcpy(&(del)->args[(argn)], &(__typeof__(argT)){ (arg) }, sizeof argT); \
	   (del)->types[(argn)] = ffi_type(__typeof__(argT)); })

#define d_setarg(del, argn, arg, ...) \
	d_copyarg((del), (argn), ffi_typeof((arg)), (arg))

#define d_setarg0(...)
#define d_setarg1(del, ...) d_setarg((del), 0, D_ARG(0, __VA_ARGS__)); d_setarg0((del), __VA_ARGS__)
#define d_setarg2(del, ...) d_setarg((del), 1, D_ARG(1, __VA_ARGS__)); d_setarg1((del), __VA_ARGS__)
#define d_setarg3(del, ...) d_setarg((del), 2, D_ARG(2, __VA_ARGS__)); d_setarg2((del), __VA_ARGS__)
#define d_setarg4(del, ...) d_setarg((del), 3, D_ARG(3, __VA_ARGS__)); d_setarg3((del), __VA_ARGS__)
#define d_setarg5(del, ...) d_setarg((del), 4, D_ARG(4, __VA_ARGS__)); d_setarg4((del), __VA_ARGS__)
#define d_setarg6(del, ...) d_setarg((del), 5, D_ARG(5, __VA_ARGS__)); d_setarg5((del), __VA_ARGS__)
#define d_setarg7(del, ...) d_setarg((del), 6, D_ARG(6, __VA_ARGS__)); d_setarg6((del), __VA_ARGS__)

#define d_setargs(del, argc, ...) \
	D_CALL(D_XPASTE(d_setarg, argc), (del), __VA_ARGS__)

#define D_GENARG0()
#define D_GENARG1() 0
#define D_GENARG2() 0, D_GENARG1()
#define D_GENARG3() 0, D_GENARG2()
#define D_GENARG4() 0, D_GENARG3()
#define D_GENARG5() 0, D_GENARG4()
#define D_GENARG6() 0, D_GENARG5()
#define D_GENARG7() 0, D_GENARG6()
#define D_GENARGS(N) D_CALL(D_XPASTE(D_GENARG, N))

#define delegate_(argc, del, fn, ...) \
	({ d_setargs(del, argc, __VA_ARGS__); \
	   delegate_init((del), ffi_type(ffi_typeof((fn)(D_GENARGS(argc)))), FFI_FN((fn)), argc); })

#define delegate(...) delegate_(D_DEC(D_DEC(D_NARG(__VA_ARGS__))), __VA_ARGS__, 0)


static inline struct delegate *delegate_init(struct delegate *del, ffi_type *rtype, void (*fn)(), int argc) {
	if (FFI_OK != ffi_prep_cif(&del->cif, FFI_DEFAULT_ABI, argc, rtype, del->types))
		return 0;

	del->fn    = fn;
	del->rtype = rtype;
	del->argc  = argc;

	return del;
} /* delegate_init() */


static inline ffi_arg invoke(struct delegate *del) {
	void *args[sizeof del->args / sizeof del->args[0]];
	ffi_arg retval;
	int argc;

	for (argc = 0; argc < del->argc; argc++)
		args[argc] = &del->args[argc];

	ffi_call(&del->cif, FFI_FN(del->fn), &retval, args);

	return retval;
} /* invoke() */


#include <event.h>

#include <sys/time.h>


#define EVENT_DELEGATE_INITIALIZER { -1, 0 }

struct event_delegate {
	int fd;
	short events;
	struct timeval timeout;
	int *uninstall, *reinstall;

	struct event event, timer;

	void (*cb)(int, short, void *);
	void *arg;

	struct delegate del;
}; /* struct event_delegate */


#define event_delegate_(argc, evd, fd, events, timeout, fn, ...) \
	({ __builtin_choose_expr( \
	   	__builtin_types_compatible_p(void (*)(int, short, void *), __typeof__(fn)), \
	   	({ (evd)->cb = (void *)fn; (evd)->arg = (void *)D_ARG(0, __VA_ARGS__); }), \
	   	({ d_setargs(&(evd)->del, argc, __VA_ARGS__); \
	   	   delegate_init(&(evd)->del, &ffi_type_void, FFI_FN(fn), argc); \
	   	   (evd)->cb = 0; (evd)->arg = 0; })); \
	   event_delegate_add((evd), (fd), (events), (timeout)); })

#define event_delegate(evd, fd, events, timeout, ...) event_delegate_(D_DEC(D_NARG(__VA_ARGS__)), (evd), (fd), (events), (timeout), __VA_ARGS__, 0)


int event_delegate_add(struct event_delegate *, int, short, struct timeval *);

void event_delegate_del(struct event_delegate *);


_Bool event_delegate_pending(struct event_delegate *, short, struct timeval *);


static __inline__ struct event_delegate *event_delegate_init(struct event_delegate *del)
	{ *del = (struct event_delegate)EVENT_DELEGATE_INITIALIZER; return del; }

static __inline__ struct event_delegate *event_delegate_reset(struct event_delegate *del)
	{ event_delegate_del(del); return event_delegate_init(del); }


#endif /* DELEGATE_H */

